from __future__ import annotations

import calendar
from collections import namedtuple
from datetime import (
    datetime,
    timedelta)
from dateutil import parser
from dateutil.parser import ParserError
from functools import partial
import logging
from matplotlib import (
    lines as mlines,
    patches as mpatches,
    pyplot as plt,
    rcParams)
from matplotlib.collections import (
    PatchCollection,
    PathCollection
)
from matplotlib.figure import Figure
from matplotlib.legend import Legend
import os
import numpy as np
import pandas as pd
from PyQt5.QtGui import QGuiApplication
from typing import TYPE_CHECKING
import warnings

from ccrenew import Numeric
from ccrenew.dashboard import daylight

# Conditional import for typechecking
if TYPE_CHECKING:
    from ccrenew.dashboard import Project

plt.ion()  # turn on interactive plots
plt.style.use('ggplot')

# Create logger
# logger = logging.getLogger(__name__)

# Create namedtuple for storing plots
Plot = namedtuple('Plot', 'fig ax')
    
# Create dict for interactive legend
legend_type_dict = {mlines.Line2D: 'lines',
                    mpatches.Rectangle: 'backgrounds',
                    PathCollection: 'markers'}

def picker(event, fig, plot_paths):
    # Assign the legend line that was clicked
    legend_icon = event.artist

    if type(legend_icon) != Legend:
        legend_label = legend_icon._label
        path_type = legend_type_dict[type(legend_icon)]

        # Find the plot line based on the legend line name
        plot_artists = plot_paths[legend_label][path_type]

        # Get visibility of the lines/markers/backgrounds & swap it
        visible = True
        for artist in plot_artists:
            visible = artist.get_visible()
            artist.set_visible(not visible)

        if path_type != 'backgrounds':
            # Make inactive lines mostly transparent in the legend
            legend_icon.set_alpha(0.2 if visible else 1.0)
        else:
            # Make inactive background color invisible on legend
            legend_icon.set_alpha(0 if visible else 0.3)

        fig.canvas.draw_idle()
        fig.canvas.flush_events()


class Plotter:
    def __init__(self, project:Project, neighbor_sensor_data:dict = {}):
        self.project = project
        self.neighbor_sensor_data = neighbor_sensor_data
        self.plot_list = {}

    def __repr__(self):
        return 'Plotter object for {}'.format(self.project.project_name)

    def __str__(self):
        return 'Plotter object for {}'.format(self.project.project_name)


    def _update_plotter_neighbors(self, neighbor_sensor_data: dict):
        self.neighbor_sensor_data = neighbor_sensor_data


    def draw_plots(self, plot_order: list|str, *args, **kwargs):
        """
        Private method to orchestrate drawing of plots

        Args:
            plotter (Plotter): [Plotter][ccrenew.dashboard.plotting.plots.Plotter] to use to draw plots for analysis
            plot_order (list): Order of plots to draw.
        """
        mth = kwargs.get('mth', None)
        close_plots = kwargs.get('close_plots', True)
        min_date = kwargs.get('min_date', None)
        max_date = kwargs.get('max_date', None)
        poa_onboarding = kwargs.get('poa_onboarding', False)
        open_folder = kwargs.get('open_folder', False)
        fullscreen = kwargs.get('fullscreen', True)
        persist = kwargs.get('persist', False)
            
        # Close all existing plots if requested
        if close_plots:
            self.close_plots(close_plots)

        # Default month to a month previous to today
        if not mth:
            mth = datetime.now().month - 1
            
        def parse_date(dt, direction, days=None):
            if isinstance(dt, Numeric):
                if direction == 'backward':
                    dt = datetime.today() - timedelta(days=abs(dt))
                else:
                    dt = datetime.today() + timedelta(days=abs(dt))
            else:
                try:
                    dt = parser.parse(dt)
                except (ValueError, ParserError):
                    if direction == 'backward':
                        days = 45 if not days else days
                        dt = datetime.today() - timedelta(days)
                    else:
                        days = 4 if not days else days
                        dt = datetime.today() + timedelta(days)
                except TypeError:
                    pass

            return dt

        # Set min & max dates for plots
        # If no min_date supplied we'll just leave it as None to show the whole plot
        if min_date:
            min_date = parse_date(min_date, 'backward')

        # Default max_date to 4 days from today to give some padding to the right side of the plot
        if max_date:
            max_date = parse_date(max_date, 'forward', 0)
        else:
            max_date = parse_date(datetime.today(), 'forward')

        # Set default date format parameters
        rcParams["date.autoformatter.day"] = "%#m/%#d/%Y"
        rcParams["date.autoformatter.hour"] = "%#m/%#d %H:%M"
        rcParams['figure.titlesize'] = 'xx-large'
        rcParams['figure.titleweight'] = 'bold'

        plot_params = {'mth': mth,
                       'min_date': min_date,
                       'max_date': max_date,
                       'poa_onboarding': poa_onboarding,
                       'fullscreen': fullscreen}

        # Display plots
        if not plot_order:
            plot_order = [
                'xplot_pwr_poa',
                'xplot_temp',
                'temps',
                'inv',
                'pr',
                '8760',
                'weather',
                'mtr_corrected',
                'mtr_dif',
                'poas',
                'pwr_poa',
                'ghi',
                'irrad',
                'tz',
                'losses',
                'poa_corr',
            ]
        
            if self.project.Battery_AC_site:
                plot_order.append('battery')
            if self.project.Tracker_site:
                plot_order.append('tracker')

        if isinstance(plot_order, str):
            plot_order = [plot_order]

        for plot_name in plot_order:
            self.plot_list[plot_name] = self._draw_plot(plot_name, plot_params, **kwargs)

        if persist:
            plt.show(block=True)
        else:
            plt.show()

        if open_folder:
            os.startfile(os.path.join(self.project.dashboard_dir,
                                      self.project.project_directory,
                                      self.project.project_name))


    def _draw_plot(self, plot_name, plot_params, **kwargs):
        # Assign parameters
        mth = plot_params['mth']
        min_date = plot_params['min_date']
        max_date = plot_params['max_date']
        poa_onboarding = plot_params['poa_onboarding']
        fullscreen = plot_params['fullscreen']

        default_tool = kwargs.get('default_tool', 'zoom')
        screen = kwargs.get('screen', None)
        tight_layout = kwargs.get('tight_layout', True)

        # Initialize fig & ax variables
        fig, ax, fig_list = None, None, None

        # Closure for drawing multiple axes on a single figure
        def draw_multiple_axes(dfs, axs, min_date, max_date):
            for df, ax in zip(dfs, axs):
                try:
                    df.plot(ax=ax)
                    ax.set_xlim(left=min_date, right=max_date)

                    # Create interactive legend
                    legend = ax.legend()
                    legend.set_draggable(True)
                    plot_paths = {line._label: {'lines': [line]} for line in ax.get_lines()}

                    for legend_handle in legend.legendHandles:
                        legend_handle.set_picker(5)

                    on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
                    fig.canvas.mpl_connect('pick_event', on_pick)
                except TypeError:
                    continue


        if plot_name.replace(' ', '_').lower() == 'xplot_pwr_poa':
            # Power vs POA Crossplot
            # We'll start by preparing the data into numpy arrays
            y = self.project.df_Pvsyst.loc[(self.project.df_Pvsyst.index.month == mth), ['Year 0 Actual Production (kWh)']].values
            x = self.project.df_Pvsyst.loc[(self.project.df_Pvsyst.index.month == mth), ['POA (W/m2)']].values
            y1 = self.project.df.loc[(self.project.df.index.month == mth), ['Meter_Corrected_2']].values
            x1 = self.project.df.loc[(self.project.df.index.month == mth), ['POA_avg']].values

            # Then create our figure & axes objects
            fig, ax = plt.subplots(num='Power vs POA - {}'.format(self.project.project_name))    
            fig.suptitle('{}\nMonth: '.format(self.project.project_name) + calendar.month_name[mth],fontsize=14, fontweight='bold')

            # Finally, we'll plot the data
            if self.project.Battery_DC_site:
                title = ('{} Power vs POA\nBattery DC Site'.format(self.project.project_name))
            elif self.project.Battery_AC_site:
                title = ('{} Power vs POA\nBattery AC Site'.format(self.project.project_name))
            else:
                title = ('{} Power vs POA'.format(self.project.project_name))
            ax.set_title(title)
            ax.scatter(x, y, c='b', label='Pvsyst [kWh/POA]')
            ax.scatter(x1, y1, c= 'r', label='Meter [kWh/POA]')
            ax.set_xlabel('POA W/m2')
            ax.set_ylabel('Power Generation kW')
            
            ax.legend()


        elif plot_name.replace(' ', '_').lower() == 'pwr_poa':
            # Power & POA Timeseries Plot
            # The following plots we'll use a slightly different convention
            # Instead of `ax.plot()` we'll use the pandas pyplot wrapper `df.plot()`
            # In our data preparation stage, if we create a pandas Series object, we'll set the name, which will pass through to the label & legend in pyplot
            # QUESTION: looks like we have two separate y-axis limits, which one do we want?
            y_orig_meter = self.project.df_Meter_ORIGINAL.loc[:,self.project.pos_Meter_ORIGINAL].sum(axis=1)
            y_orig_meter.name = 'Original'
            y_meter_corrected = self.project.df.loc[:, 'Meter_Corrected_2']
            y_meter_corrected.name = 'Meter Corrected'
            y_inv_sum = self.project.ava.loc[:, 'Inv_sum']
            y_inv_sum.name = 'Inv Sum'
            y_poa = self.project.df.loc[:, 'POA_avg']
            y_poa.name = 'POA Avg'
            mc_max:float = y_meter_corrected.max()
            meter_corrected_ylim = (-0.05*mc_max, mc_max/0.9)
            poa_max:float = y_poa.max()
            poa_ylim = (-0.05*poa_max, poa_max/0.9)

            fig, ax = plt.subplots(num='Power & POA - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)
            ax_power = ax
            ax_poa = ax_power.twinx()

            if self.project.Battery_DC_site:
                title = 'Power & POA\nBattery DC Site'.format(self.project.project_name)
            elif self.project.Battery_AC_site:
                title = 'Power & POA\nBattery AC Site'.format(self.project.project_name)
            elif self.project.df_proj_keys['Fund'] == 'USB 1':
                title = 'Power & POA\n***NOTE: USB site - will need a utility invoice for meter data***'.format(self.project.project_name)
            else:
                title = 'Power & POA'.format(self.project.project_name)
            ax.set_title(title)
            y_orig_meter.plot(ax=ax_power)
            y_meter_corrected.plot(ax=ax_power, linestyle=':', linewidth=5)
            y_inv_sum.plot(ax=ax_power, linestyle='-.')
            ax.axhline(y = self.project.MWAC * 1000, color = 'r', linestyle = '--', label='Clipping Limit')
            y_poa.plot(ax=ax_poa, color='grey')
            ax_power.set_xlim(left=min_date, right=max_date)
            ax_power.set_ylim(meter_corrected_ylim)
            ax_poa.set_ylim(poa_ylim)

            # Create interactive legend
            legend = fig.legend()
            legend.set_draggable(True)
            plot_paths = {line.get_label(): {'lines': [line]} for line in list(ax_power.get_lines()) + list(ax_poa.get_lines())}

            for legend_handle in legend.legendHandles:
                legend_handle.set_picker(5)

            on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
            fig.canvas.mpl_connect('pick_event', on_pick)


        elif plot_name.replace(' ', '_').lower() == 'mtr_corrected':
            # Meter Corrections
            y_orig_meter = self.project.df_Meter_ORIGINAL.loc[:,self.project.pos_Meter_ORIGINAL].sum(axis=1)
            y_orig_meter.name = 'Original'
            y_meter_corrected = self.project.df.loc[:, 'Meter_Corrected_2']
            y_meter_corrected.name = 'Meter Corrected'
            y_inv_sum = self.project.ava.loc[:, 'Inv_sum']
            y_inv_sum.name = 'Inv Sum'
            y_cum_orig = self.project.df_Meter_ORIGINAL.loc[:,self.project.pos_Meter_Cum_ORIGINAL].sum(axis=1)
            y_cum_orig.name = 'Original'
            y_cum_corrected = self.project.df.loc[:, 'Meter_cum_corrected_2']
            y_cum_corrected.name = 'Meter Corrected'
            inv_offset = self.project.df[self.project.pos_Inv_cum].sum(axis=1).subtract(self.project.df_Meter_ORIGINAL.loc[:,self.project.pos_Meter_Cum_ORIGINAL].sum(axis=1)).median()
            y_cum_inv = self.project.df[self.project.pos_Inv_cum].sum(axis=1).subtract(inv_offset)
            y_cum_inv.name = 'Inv Sum'   

            fig, axs = plt.subplots(2, num='Meter Corrections - {}'.format(self.project.project_name), sharex=True)
            fig.suptitle(self.project.project_name)
            ax_power = axs[0]
            ax_cum = axs[1]
            
            if self.project.Battery_DC_site:
                title = 'Power Meter Correction\nBattery DC Site'
            elif self.project.Battery_AC_site:
                title = 'Power Meter Correction\nBattery AC Site'
            elif self.project.df_proj_keys['Fund'] == 'USB 1':
                title = 'Power Meter Correction\n***NOTE: USB site - will need a utility invoice for meter data***'
            else:
                title = ' Power Meter Correction'
            ax_power.set_title(title)
            y_orig_meter.plot(ax=ax_power)
            y_meter_corrected.plot(ax=ax_power, linestyle=':',linewidth=5)
            y_inv_sum.plot(ax=ax_power, linestyle='-.')
            ax_power.axhline(y = self.project.MWAC * 1000, color = 'r', linestyle = '--', label='Clipping Limit')
            ax_power.set_xlim(left=min_date, right=max_date)
            ax_power.set_ylim((-100, self.project.MWAC * 1.3 * 1000))
            ax_power.legend()

            ax_cum.set_title('Cumulative Energy Meter Correction')
            y_cum_orig.plot(ax=ax_cum)
            y_cum_corrected.plot(ax=ax_cum)
            y_cum_inv.plot(ax=ax_cum, linestyle='-.')
            ax_cum.set_xlim(left=min_date, right=max_date)
            ax_cum.legend()

            # We'll turn the x-axis labels back on for the top plot because pyplot disables those by default when you share an axis
            ax_power.xaxis.set_tick_params(labelbottom=True, which='both')


        elif plot_name.replace(' ', '_').lower() == 'poas':
            # POA sensors
            y_poa = self.project.df.loc[:, self.project.pos_POA]
            y_poa_solcast = self.project.df_solcast.loc[:, 'POA_avg']
            y_poa_solcast.index.name = None
            y_poa_solcast.name = 'Solcast'
            y_poa_avg = self.project.df.loc[:, 'POA_avg']
            y_poa_avg.name = 'POA Avg'
            y_neighbors = pd.DataFrame()
            for neighbor_name, neighbor_df in self.neighbor_sensor_data.items():
                if self.project.neighbor_list[neighbor_name]:
                    y_neighbors[neighbor_name] = neighbor_df.loc[:, 'POA_avg']
            racking = self.project.df_proj_keys.get('Racking')
            
            fig, ax = plt.subplots(num='POAs - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            title = 'Irradiance - Racking: {}'.format(racking)
            if self.project.df_proj_keys['Fund'] == 'Soltage Landfill':
                title = 'Irradiance - Racking: {}\n***NOTE: POAs might be off tilt from each other due to site geography. The average may still be acceptable.***'.format(racking)
            ax.set_title(title)

            # We'll use the default colormap but remove '#FBC15E' to use as the Solcast color
            cmap = plt.rcParams['axes.prop_cycle'].by_key()['color']
            solcast_color = '#FBC15E'
            cmap.remove(solcast_color)
            ax.set_prop_cycle(color=cmap)

            # Plot data
            y_poa.plot(ax=ax)
            if not y_neighbors.empty:
                y_neighbors.plot(ax=ax)
            y_poa_solcast.plot(ax=ax, linestyle='--', color=solcast_color)

            ax.set_xlim(left=min_date, right=max_date)

            # Initialize dictionary for interactive legend
            plot_paths = {line.get_label(): {'lines': [line]} for line in ax.get_lines()}

            # Group POA_avg data sources to plot as scatter markers
            poa_source_df_dict = self.project.df[['POA_avg', 'POA_source']].groupby('POA_source').groups

            # Order sources to show Native, Config, Solcast followed by others
            poa_source_df_dict_sorted = {}
            poa_source_list = ['Native', 'Config', 'Config MT', 'Solcast', 'Solcast MT']
            for source in poa_source_list:
                try:
                    poa_source_df_dict_sorted[source] = poa_source_df_dict.pop(source)
                except KeyError:
                    continue

            # Sort the neighbors to show as closest first in order to farthest
            # (project.neighbor_list is already in this order, we'll use that as a key for the sorted function)
            neighbors_sorted = list(self.project.neighbor_list.keys())
            for neighbor in neighbors_sorted:
                # Add neighbors
                try:
                    poa_source_df_dict_sorted[neighbor] = poa_source_df_dict.pop(neighbor)
                except KeyError:
                    pass
                try:
                    poa_source_df_dict_sorted[neighbor + ' MT'] = poa_source_df_dict.pop(neighbor + ' MT')
                except KeyError:
                    continue

            # Custom markers for data source type
            marker_dict = {'Native': 'o', 'Config': '$c$', 'Config MT': '$c$', 'Solcast': '*', 'Solcast MT': '*'}
            neighbor_markers = ['x', '+', 'd', '^', '2', '<', 's', 'p', '>', 'h', 'H', 'D', 'P', 'v', (6,2,90)]
            background_color_dict = {'Native': 'dodgerblue', 'Config': 'palegreen', 'Config MT': 'palegreen', 'Solcast': 'gold', 'Solcast MT': 'gold'}
            for i, neighbor in enumerate(neighbors_sorted):
                marker_dict[neighbor] = neighbor_markers[i]
                marker_dict[neighbor + ' MT'] = neighbor_markers[i]

            # Plot the data source points
            background_handles = []
            for i, (source, idx) in enumerate(poa_source_df_dict_sorted.items()):
                # There's a weird bug in the ax.scatter method that errors out if there's only one x-axis Pandas datetime point
                # So when there's only one point we'll just add itself to the index & carry on
                if len(idx) == 1:
                    idx = idx.append(idx)
                if ' MT' in source:
                    color = 'black'
                else:
                    color = 'red'
                markers = ax.scatter(x=idx, y=y_poa_avg[idx], marker=marker_dict[source], label=source, color=color)
                try:
                    plot_paths[source]['markers'] = [markers]
                except KeyError:
                    plot_paths[source] = {'markers': [markers]}

                # Add background color
                source_spans = pd.DataFrame({'data': idx})
                source_spans['span_starts'] = source_spans['data'].diff() != pd.Timedelta(hours=1)
                source_spans['span_ends'] = source_spans['data'].diff(-1) != pd.Timedelta(days=-1, hours=23)
                span_starts = source_spans['data'][source_spans['span_starts']]
                span_ends = source_spans['data'][source_spans['span_ends']] + pd.Timedelta(hours=1)

                # Standard colors for Native/Config/Solcast, violet for neighbors
                try:
                    background_color = background_color_dict[source]
                except KeyError:
                    background_color = 'violet'

                # Create patch collection for each data source
                patches = []
                for start, end in zip(span_starts, span_ends):
                    xmin, xmax = start, end
                    ymin, ymax = 0, 1

                    # Convert from dates to floats
                    (xmin, xmax), = ax._process_unit_info([('x', [xmin, xmax])])

                    verts = [(xmin, ymin), (xmin, ymax), (xmax, ymax), (xmax, ymin)]
                    p = mpatches.Polygon(verts, color=background_color, alpha=0.3, label=source)

                    patches.append(p)
                    
                patch_collection = PatchCollection(
                        patches,
                        facecolors=background_color,
                        edgecolors=background_color,
                        alpha=0.3,
                        label=source)

                # We need to first add the collection to the axes then apply the transform lest everything get all wonky
                ax.add_collection(patch_collection)
                background = ax.collections[-1]
                background.set_transform(ax.get_xaxis_transform(which='grid'))

                # Since Matplot lib (version 3.5.2) doesn't allow for creation of a legend entry for patches
                # We'll just choose one patch to represent the collection in the legend
                # label_stripped = source.replace(' MT', '')
                # if label_stripped not in background_labels:
                #     background_labels.append(label_stripped)

                # Update plot_paths with the background
                label_stripped = source.replace(' MT', '')
                try:
                    plot_paths[label_stripped]['backgrounds'].append(background)
                except KeyError:
                    plot_paths[label_stripped]['backgrounds'] = [background]
                    background_handles.append(p)

            # Create interactive legend
            # We'll need to exclude the background PatchCollections then add the individual patches in the next step due to the issue above
            legend_handles = [plot_artist[0] for artist_dict in plot_paths.values() for plot_type, plot_artist in artist_dict.items() if plot_type != 'backgrounds']
            legend_handles.extend(background_handles)
            visible_lines = self.project.col_ref['POA_cols'] + list(poa_source_df_dict_sorted)
            
            legend = ax.legend(handles=legend_handles)
            legend.set_draggable(True)

            # Add Solcast to the list of visible lines if it's only been added due to mistracking
            if 'Solcast MT' in visible_lines and 'Solcast' not in visible_lines:
                visible_lines.append('Solcast')
            path_dict = {}

            for legend_handle in legend.legendHandles:
                source = legend_handle._label
                path_dict = plot_paths[source]
                legend_handle.set_picker(True)

                # For line-type legend icons we'll add some padding around the line so you don't have to click exactly on the line
                # Line padding in pixels
                line_padding = 5

                for path_type, paths in path_dict.items():
                    if path_type == 'lines':
                        legend_handle.set_picker(line_padding)

                        # Turn line off if not used for data sub
                        if source not in visible_lines:
                            for path in paths:
                                path.set_visible(False)
                            legend_handle.set_alpha(0.2)

                    elif path_type == 'markers':
                        # Bring marker points in front of all lines
                        for path in paths:
                            path.set_zorder(100)

                    else:
                        # Push backgrounds to the back of the pot
                        for path in paths:
                            path.set_zorder(0)


            on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
            fig.canvas.mpl_connect('pick_event', on_pick)


        elif plot_name.replace(' ', '_').lower() == 'poa_corr':
            # POA Correlations
            if len(self.project.pos_POA) > 1:
                if poa_onboarding:
                    # Plot all months
                    y = self.project.df.loc[:,self.project.pos_POA]
                else:
                    # Plot only selected month
                    y = self.project.df.loc[self.project.df.index.month == mth, self.project.pos_POA]
                
                y.set_index(y.columns[0], drop=False, inplace=True)

                fig, ax = plt.subplots(num='POA Correlations - {}'.format(self.project.project_name))
                fig.suptitle(self.project.project_name)

                title='POA Correlation Check'
                if self.project.df_proj_keys['Fund'] == 'Soltage Landfill':
                    title='POA Correlation Check\n***NOTE: POA correlation may show hysteresis due to site geography***'
                ax.set_title(title)
                y.plot(ax=ax, marker='o', linestyle='None')
                ax.set_xlim(-10,1300)
                ax.set_ylim(-10,1300)

                # Create interactive legend
                legend = ax.legend()
                lines =  [line for line in ax.get_lines()]
                path_dict = {}

                for leg_line, plot_line in zip(legend.get_lines(), lines):
                    leg_line.set_picker(5)
                    path_dict[leg_line] = plot_line

                on_pick = partial(picker, fig=fig, path_dict=path_dict)
                fig.canvas.mpl_connect('pick_event', on_pick)
                
            else:
                print('Only one POA! No plot for correlation')


        elif plot_name.replace(' ', '_').lower() == 'xplot_poa_sol':
            # POA sensors
            poa_cols = self.project.col_ref['POA_cols']
            if not min_date:
                min_date = self.project.df.index[0]
            df_poa = self.project.df.query("index>=@min_date and index<@max_date").reindex(columns=poa_cols)
            df_poa = daylight(lambda **kwargs: kwargs['df'])(df=df_poa, project=self.project)
            native_poa_cols = df_poa.columns
            df_poa['Solcast'] = self.project.df_solcast.loc[:, 'POA_avg']
            x = df_poa.loc[:, 'Solcast']
            y = df_poa.loc[:, native_poa_cols]

            racking = self.project.df_proj_keys.get('Racking')
            
            fig, ax = plt.subplots(num='POA vs Solcast - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            title = 'POA vs Solcast - Racking: {}'.format(racking)
            if self.project.df_proj_keys['Fund'] == 'Soltage Landfill':
                title = 'Irradiance - Racking: {}\n***NOTE: POAs might be off tilt from each other due to site geography. The average may still be acceptable.***'.format(racking)
            ax.set_title(title)

            markers = ['o', '+', '^', 'x', 's', '<', '*', 'd', 'v', '2', '>', '$D$']

            # We'll use the default colormap but remove '#FBC15E' to use as the Solcast color
            cmap = plt.rcParams['axes.prop_cycle'].by_key()['color']
            cmap = cmap[:len(native_poa_cols)]
            ax.set_prop_cycle(color=cmap)

            df_months = df_poa.groupby(df_poa.index.month)
            for i, (month, df) in enumerate(df_months):
                x = df['Solcast']
                y = np.array(df.loc[:, native_poa_cols])
                ax.plot(x, y, marker = markers[i], linestyle='None')
                # for j, col in enumerate(native_poa_cols):
                #     x = df['Solcast']
                #     y = df.loc[:, col]
                #     label = f"{calendar.month_abbr[month]} {col}"
                #     ax.plot(x, y, c=colors[j], marker=markers[i], label=label)

            # Create interactive legend
            # labels = [f"{calendar.month_abbr[month]} {col}" for month, _ in df_months for col in native_poa_cols]
            # legend = ax.legend(labels)
            # lines =  [line for line in ax.collections]
            # path_dict = {}

            # for leg_line, plot_line in zip(legend.legendHandles, lines):
            #     leg_line.set_picker(True)
            #     path_dict[leg_line] = plot_line

            # on_pick = partial(picker, fig=fig, path_dict=path_dict)
            # fig.canvas.mpl_connect('pick_event', on_pick)

        
        elif plot_name.replace(' ', '_').lower() == 'inv':
            # Inverters
            y = self.project.df.loc[:, self.project.pos_Inv]

            fig, ax = plt.subplots(num='Inverters - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            ax.set_title('Inverter KW')
            y.plot(ax=ax)
            ax.set_xlim(left=min_date, right=max_date)

            # Create interactive legend
            legend = ax.legend()
            legend.set_draggable(True)
            plot_paths = {line._label: {'lines': [line]} for line in ax.get_lines()}

            for legend_handle in legend.legendHandles:
                legend_handle.set_picker(5)

            on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
            fig.canvas.mpl_connect('pick_event', on_pick)

            # TODO: is this for determining inverter outages?
            def list_pairs(lis):
                #find pairs that are at least 2 hours in duration
                beg = []
                end = []
                            
                start_time = 0
                
                while start_time < len(lis):
                    end_time = start_time+1
                    while end_time < len(lis) and lis[end_time] - lis[end_time-1] == pd.Timedelta('1h'):
                        end_time = end_time+1
                    # end_time overshoots by one hour, so adjust to make sure it's longer than 1 hr
                    if (end_time-1) - start_time >= 1:
                        beg.append(lis[start_time])
                        end.append(lis[end_time-1])
                    start_time = end_time
                return beg, end
                
            beg, end = list_pairs(list(self.project.df.loc[self.project.df['Meter_&_ava'] > self.project.df['Meter_Corrected_2'], :].index) )

            for beg,end in zip(beg,end):
                ax.axvspan(beg, end, facecolor='#2ca02c', alpha=0.5)


        elif plot_name.replace(' ', '_').lower() == 'pr':
            # Hourly PR
            freq = 'h'
            top = self.project.df.loc[:, 'Gen_NO_Clipping'].resample(freq).sum()
            bottom = self.project.df.loc[:, 'DC_corrected_PR'].resample(freq).sum()
            y = top.div(bottom)
            y_avg = y.dropna().rolling(48).mean()

            fig, ax = plt.subplots(num='PR - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            ax.set_title('Hourly PR')
            y.plot(ax=ax, marker='o', linestyle='None', label='Hourly PR')
            try:
                y_avg.plot(ax=ax, label='48 Hour Rolling Avg')
            except Exception as e:
                print('Skipped adding PR line for {}'.format(self.project.project_name))
                # logger.warn('{} error plotting PR rolling avg for {}'.format(e, self.project.project_name))
            ax.set_xlim(left=min_date, right=max_date)
            
            # Create interactive legend
            legend = ax.legend()
            legend.set_draggable(True)
            plot_paths = {line._label: {'lines': [line]} for line in ax.get_lines()}

            for legend_handle in legend.legendHandles:
                legend_handle.set_picker(5)

            on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
            fig.canvas.mpl_connect('pick_event', on_pick)
        

        elif plot_name.replace(' ', '_').lower() == 'losses':
            # Losses
            y_losses = self.project.losses

            # If single inverter site, remove losses where meter_corrected_2 > 0
            if len(self.project.pos_Inv) == 1:
                y_losses.loc[(y_losses['Meter_Corrected_2'] > 0), 'Inv_losses'] = 0
            
            # Set negative losses to zero
            y_losses.clip(lower=0, inplace=True)

            fig, ax = plt.subplots(num = 'Losses - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            ax.set_title('Losses by Type')
            y_losses.plot.area(ax=ax, linewidth=0)
            ax.set_xlim(left=min_date, right=max_date)

            # Assign column labels to lines because Pandas `plot.area()` doesn't do that for you automagically like `plot()` does
            for (ln, col) in zip(ax.lines, y_losses.columns):
                ln.set_label(col)


        elif plot_name.replace(' ', '_').lower() == 'temps':
            # Temperatures
            y_tcell = self.project.df['Tcell']
            y = self.project.df[['Tamb_avg', 'Tcell_AMB', 'Tcell_MOD']]
            
            # Add Fahrenheit conversion
            y_tcell_fahr = y_tcell*1.8+32
            y_fahr = y*1.8+32

            fig, ax = plt.subplots(num='Temperatures - {}'.format(self.project.project_name))
            ax_cels = ax
            ax_fahr = ax_cels.twinx()
            fig.suptitle(self.project.project_name)

            project_state = self.project.df_proj_keys['State']
            ax_cels.set_title(f'Temperatures - {project_state}')
            y_tcell.plot(ax=ax_cels, marker='o', color='r')
            y.plot(ax=ax_cels)
            y_tcell_fahr.plot(ax=ax_fahr, style='None', grid=False)
            y_fahr.plot(ax=ax_fahr, style='None', grid=False)
            ax_cels.set_ylabel('Temperature (C)')
            ax_fahr.set_ylabel('Temperature (F)')
            ax_cels.set_xlim(left=min_date, right=max_date)

            # Create interactive legend
            legend = ax_cels.legend()
            legend.set_draggable(True)
            plot_paths = {line._label: {'lines': [line]} for line in ax_cels.get_lines()}

            for legend_handle in legend.legendHandles:
                legend_handle.set_picker(5)

            on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
            fig.canvas.mpl_connect('pick_event', on_pick)

            ax_fahr.legend().set_visible(False)


        elif plot_name.replace(' ', '_').lower() == 'weather':
            # Weather Sensors
            y_poa = self.project.df['POA_avg']
            y_tamb = self.project.df['Tamb_avg']
            y_tmod = self.project.df['Tmod_avg']
            y_wind = self.project.df['Wind_speed']

            fig, axs = plt.subplots(4, num='Weather Sensors - {}'.format(self.project.project_name), sharex=True)
            fig.suptitle(self.project.project_name)
            ax_poa = axs[0]
            ax_tamb = axs[1]
            ax_tmod = axs[2]
            ax_wind = axs[3]

            ax_poa.set_title('Irradiance')
            y_poa.plot(ax=ax_poa, label='POA Avg')
            ax_tamb.set_title('Ambient Temperature')
            y_tamb.plot(ax=ax_tamb, label='Tamb Avg')
            ax_tmod.set_title('Ambient Temperature')
            y_tmod.plot(ax=ax_tmod, label='Tamb Avg')
            ax_wind.set_title('Wind Speed')
            y_wind.plot(ax=ax_wind, label='Wind Speed')
            for ax in axs:
                ax.set_xlim(left=min_date, right=max_date)
                
                # Create interactive legend
                legend = ax.legend()


        elif plot_name.replace(' ', '_').lower() == 'ghi':
            #GHI plot
            if len(self.project.pos_GHI) > 0:
                y_ghi = self.project.df.loc[:, self.project.pos_GHI]
                y_ghi_avg = self.project.df.loc[:, 'GHI_avg']
                y_poa = self.project.df.loc[:, self.project.pos_POA]
                y_poa_avg = self.project.df.loc[:, 'POA_avg']
                
                try:
                    fig, ax = plt.subplots(num = 'GHI - {}'.format(self.project.project_name))
                    fig.suptitle(self.project.project_name)
                
                    ax.set_title('GHI Sensors')
                    y_ghi.plot(ax=ax)
                    y_ghi_avg.plot(ax=ax, style='o', label='GHI Avg')
                    ax.set_xlim(left=min_date, right=max_date)
                    
                    # Create interactive legend
                    legend = ax.legend()
                    legend.set_draggable(True)
                    plot_paths = {line._label: {'lines': [line]} for line in ax.get_lines()}

                    for legend_handle in legend.legendHandles:
                        legend_handle.set_picker(5)

                    on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
                    fig.canvas.mpl_connect('pick_event', on_pick)

                except:
                    print("Couldn't make GHI_avg plot")
                    fig, ax = 'GHI plot error', 'GHI plot error'
            else:
                print("No GHI sensors to build the GHI plots")
                fig, ax = 'No GHI to plot', 'No GHI to plot'
                

        elif plot_name.replace(' ', '_').lower() == 'irrad':
            #POA & GHI Plot
            # Get GHI sensors if present
            if len(self.project.pos_GHI) > 0:
                y_ghi = self.project.df.loc[:, self.project.pos_GHI]
                y_ghi_avg = self.project.df.loc[:, 'GHI_avg']

            # Get POAs
            y_poa = self.project.df.loc[:, self.project.pos_POA]
            y_poa_avg = self.project.df.loc[:, 'POA_avg']

            try:
                fig, ax = plt.subplots(num = 'POA & GHI - {}'.format(self.project.project_name))
                fig.suptitle(self.project.project_name)
                
                ax.set_title('Irradiance Sensors')
                y_ghi.plot(ax=ax)
                y_ghi_avg.plot(ax=ax, label='GHI Avg', style='-.')
                y_poa.plot(ax=ax)
                y_poa_avg.plot(ax=ax, label='POA Avg', style=':')                               
                ax.set_xlim(left=min_date, right=max_date)
                
                # Create interactive legend
                legend = ax.legend()
                legend.set_draggable(True)
                plot_paths = {line._label: {'lines': [line]} for line in ax.get_lines()}

                for legend_handle in legend.legendHandles:
                    legend_handle.set_picker(5)

                on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
                fig.canvas.mpl_connect('pick_event', on_pick)

            except Exception as e:
                print("Couldn't make POA vs GHI plot. Exception: {}".format(e))
                fig, ax = 'POA vs GHI plot error', 'POA vs GHI plot error'
            

        elif plot_name.replace(' ', '_').lower() == 'tz':
            #Timezone plot and night check
            x = self.project.df_filt.index.hour
            y_meter = self.project.df_filt['Meter_Corrected_2']
            y_poa = self.project.df_filt['POA_avg']*4
            scale = self.project.df_filt.index.month
            interp_hours_list = list(set(self.project.df_filt.loc[self.project.df_filt['interp_check'] == 1, 'Hour Index Copy'].tolist()))

            fig, ax = plt.subplots(num='Timezone Check - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            ax.set_title('Timezone Check')
            # We'll store the scatter plot in a variable so we can reference it with the colorbar
            sp = ax.scatter(x, y_meter, c=scale, cmap ='jet')
            sp = ax.scatter(x, y_poa, c=scale, cmap ='jet')
            cb = plt.colorbar(sp)
            cb.set_label('Month')
            ax.set_xlim(0,24)

            for interp_hour in interp_hours_list:
                ax.axvspan(interp_hour - 0.25, interp_hour + 0.25, facecolor='#2ca02c', alpha=0.5)
            

        elif plot_name.replace(' ', '_').lower() == 'mtr_dif':
            # Meter Corrected Dif
            try:
                original_meter_cum = self.project.df_Meter_ORIGINAL.drop(self.project.pos_Meter_ORIGINAL,1).sum(axis=1)
                corrected_meter_cum = self.project.df['Meter_cum_corrected_2']
                y = original_meter_cum.subtract(corrected_meter_cum)
                min_y = y[(self.project.df_Meter_ORIGINAL.drop(self.project.pos_Meter_ORIGINAL,1) != 0).all(axis=1)].min()
                max_y = y[(self.project.df_Meter_ORIGINAL.drop(self.project.pos_Meter_ORIGINAL,1) != 0).all(axis=1)].max()
                
                fig, ax = plt.subplots(num='Meter Correction Dif - {}'.format(self.project.project_name))
                fig.suptitle(self.project.project_name)
                
                ax.set_title('Meter Correction Dif')
                y.plot(ax=ax)
                ax.set_xlim(left=min_date, right=max_date)
                ax.set_ylim(min_y, max_y)
            except:
                print("Couldn't build the meter corrected difference plot")
                fig, ax = 'Meter corrected dif error', 'Meter corrected dif error'
                    
            
        elif plot_name.replace(' ', '_').lower() == '8760':
            # 8760 vs Measured Power
            y_8760 = self.project.df_Pvsyst['Year 0 Actual Production (kWh)']
            y_meter_corrected = self.project.df['Meter_Corrected_2']


            fig, ax = plt.subplots(num='Measured/8760 Generation Comparison - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            ax.set_title('Measured/8760 Generation Comparison')
            y_8760.plot(ax=ax, label='8760', c='blue')
            y_meter_corrected.plot(ax=ax, label='Meter Corrected', c='red')
            
            # Create interactive legend
            legend = ax.legend()
            legend.set_draggable(True)
            plot_paths = {line._label: {'lines': [line]} for line in ax.get_lines()}

            for legend_handle in legend.legendHandles:
                legend_handle.set_picker(5)

            on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
            fig.canvas.mpl_connect('pick_event', on_pick)
            

        elif plot_name.replace(' ', '_').lower() == 'xplot_temp':
            # Power vs POA by Temperature
            x = self.project.df.loc[:, 'POA_avg']
            y = self.project.df.loc[:, 'Meter_Corrected_2']
            x_max = 1300
            y_max = self.project.df['Meter_Corrected_2'].max()*1.05
            scale = self.project.df.loc[:, 'Tamb_avg']

            fig, ax = plt.subplots(num='YTD Power vs POA, colored by ambient temp - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            ax.set_title('YTD Power vs POA, colored by ambient temp')
            # We'll store the scatter plot in a variable so we can reference it with the colorbar
            sp = ax.scatter(x, y, c=scale, cmap='jet')    
            cb = plt.colorbar(sp)
            cb.set_label('Tamb Avg')
            ax.set_xlim(0, x_max)
            ax.set_ylim(0, y_max)
            ax.set_ylabel('Production [KWh]')
            ax.set_xlabel('Irradiance [W/m2]')
        

        elif plot_name.replace(' ', '_').lower() == 'native':
            poa_cols = [col for col in self.project.pos_POA if any([col in nat for nat in self.project.pos_Native])]
            ghi_cols = [col for col in self.project.pos_GHI if any([col in nat for nat in self.project.pos_Native])]
            tamb_cols = [col for col in self.project.pos_Tamb if any([col in nat for nat in self.project.pos_Native])]
            tmod_cols = [col for col in self.project.pos_Tmod if any([col in nat for nat in self.project.pos_Native])]
            wind_cols = [col for col in self.project.pos_Wind if any([col in nat for nat in self.project.pos_Native])]
            poa = self.project.df[poa_cols]
            ghi = self.project.df[ghi_cols]
            tamb = self.project.df[tamb_cols]
            tmod = self.project.df[tmod_cols]
            wind = self.project.df[wind_cols]

            fig, axs = plt.subplots(5, sharex=True, num='Native Weather - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            draw_multiple_axes([poa, ghi, tamb, tmod, wind], axs, min_date, max_date)


        elif plot_name.replace(' ', '_').lower() == 'sensor_avg':
            poa = self.project.df['POA_avg']
            ghi = self.project.df['GHI_avg']
            tamb = self.project.df['Tamb_avg']
            tmod = self.project.df['Tmod_avg']
            wind = self.project.df['Wind_speed']

            fig, axs = plt.subplots(5, sharex=True, num='Averaged Weather - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            draw_multiple_axes([poa, ghi, tamb, tmod, wind], axs, min_date, max_date)


        elif plot_name.replace(' ', '_').lower() == 'meters':
            fig_list = []
            for i in range(len(self.project.pos_Meter)):
                meter_name = self.project.pos_Meter[i]
                y_meter_orig = self.project.df_Meter_ORIGINAL.iloc[:,i]
                y_meter_corr = self.project.df[meter_name]
                y_cum_orig = self.project.df_Meter_ORIGINAL.iloc[:,[i+len(self.project.pos_Meter)]]
                y_cum_corr = self.project.df[[meter_name]].cumsum()
                y_cum_corr = y_cum_corr.rename(columns = {meter_name: meter_name.replace('_kw_', '_kwhnet_')})
                
                # Add in a vertical offset if the original & corrected cum meters don't match up
                condition = y_cum_orig.loc[(y_cum_orig[y_cum_orig.columns[0]] > 0), :].index
                offset = np.mean(y_cum_orig.loc[condition, :].values - y_cum_corr.loc[condition, :].values)
                y_cum_corr = y_cum_corr.add(offset)

                fig, axs = plt.subplots(2, num='Meter {} - {}'.format(i+1, self.project.project_name), sharex=True)
                fig.suptitle('{}'.format(self.project.project_name))
                ax_meter = axs[0]
                ax_cum = axs[1]

                ax_meter.set_title('Meter {}'.format(i+1))
                y_meter_orig.plot(ax=ax_meter)
                y_meter_corr.plot(ax=ax_meter)
                ax_meter.set_xlim(left=min_date, right=max_date)

                y_cum_orig.plot(ax=ax_cum)
                y_cum_corr.plot(ax=ax_cum)
                ax_cum.set_xlim(left=min_date, right=max_date)

                # Create interactive legend
                meter_legend = ax_meter.legend(['Original', 'Corrected'])
                meter_legend.set_draggable(True)

                cum_legend = ax_cum.legend(['Original', 'Corrected'])
                cum_legend.set_draggable(True)

                plot_paths = {line._label: {'lines': [line]} for line in list(ax_meter.get_lines()) + list(ax_cum.get_lines())}

                for legend_handle in list(meter_legend.legendHandles) + list(cum_legend.legendHandles):
                    legend_handle.set_picker(5)

                on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
                fig.canvas.mpl_connect('pick_event', on_pick)

                fig_list.append(fig)


        elif plot_name.replace(' ', '_').lower() == 'snow':
            # Snow Plots
            y_snow_data = self.project.snow_data['snow']
            y_snow_coverage = self.project.snow_coverage

            fig, ax = plt.subplots(num='Snow Data - {}'.format(self.project.project_name))
            fig.suptitle(self.project.project_name)

            ax.set_title('Snow Data')
            y_snow_data.plot(ax=ax, label='Snowfall', c='blue')
            y_snow_coverage.plot(ax=ax, label='Coverage', c='red')
            
            
            # Create interactive legend
            legend = ax.legend()
            legend.set_draggable(True)
            plot_paths = {line._label: {'lines': [line]} for line in ax.get_lines()}

            for legend_handle in legend.legendHandles:
                legend_handle.set_picker(5)

            on_pick = partial(picker, fig=fig, plot_paths=plot_paths)
            fig.canvas.mpl_connect('pick_event', on_pick)


        elif plot_name.replace(' ', '_').lower() == 'battery':
            if self.project.Battery_AC_site:
                # Battery Discharge
                poi_cols = [s for s in self.project.df_POI_ORIGINAL.columns if 'POI_kw_' in s]
                y_poi_measured = self.project.df_POI_ORIGINAL.loc[:, poi_cols].sum(axis=1)
                y_poi_modeled = self.project.df.loc[:, 'POI_modeled']
                y_discharge = self.project.df.loc[self.project.df.Rates == self.project.df.Rates.max(), poi_cols].sum(axis=1)
                y_comparison = self.project.df.loc[:, 'POI_modeled'] - self.project.df['POI_Corrected_2']
                y_comparison.loc[self.project.df.Rates == self.project.df.Rates.min()] = 0

                fig, axs = plt.subplots(2, num='Battery Discharge - {}'.format(self.project.project_name), sharex=True)
                fig.suptitle(self.project.project_name)
                ax_discharge = axs[0]
                ax_comparison = axs[1]

                ax_discharge.set_title('Measured/Expected Battery Discharge Comparison')
                ax_discharge.plot(y_poi_measured, label='POI Measured')
                ax_discharge.plot(y_poi_modeled, label='POI Modeled')
                ax_discharge.plot(y_discharge, label='Discharge', c = 'k', ls='', marker = 'x', ms = 8)
                ax.set_xlim(left=min_date, right=max_date)
                ax_discharge.legend()

                ax_comparison.set_title('Expected/Measured Difference')
                ax_comparison.plot(y_comparison)
                            
                # POI Energy Correction
                poi_cols = [s for s in self.project.df_POI_ORIGINAL.columns if 'POI_kw_' in s]
                poi_cum_cols = [s for s in self.project.df_POI_ORIGINAL.columns if 'POI_kwhnet_' in s]
                y_poi_original = self.project.df_POI_ORIGINAL[poi_cols].sum(axis=1)
                y_poi_corrected = self.project.df.loc[:, 'POI_Corrected_2']
                y_poi_cum_original = self.project.df_POI_ORIGINAL[poi_cum_cols].sum(axis=1)
                # These next three lines adjust the corrected cum meter to match the y-intercept of the original cum meter
                poi_cum_corrected = self.project.df.loc[:, 'POI_Corrected_2'].cumsum()
                poi_cum_meter_offset = self.project.df_POI_ORIGINAL[poi_cum_cols].sum(axis=1).subtract(poi_cum_corrected).median()
                y_poi_cum_corrected = self.project.df.loc[:, 'POI_cum_corrected_2'].add(poi_cum_meter_offset)

                fig, axs = plt.subplots(2, num='POI Energy Correction - {}'.format(self.project.project_name), sharex=True)
                fig.suptitle(self.project.project_name)
                ax_correction = axs[0]
                ax_cum_correction = axs[1]

                ax_correction.set_title('POI Energy Correction')
                ax_correction.plot(y_poi_original, label='Original')  
                ax_correction.plot(y_poi_corrected, label='Corrected')
                ax_correction.axhline(y = self.project.MWAC * 1000, color = 'r', linestyle = '--', label='MWAC')
                ax_correction.set_xlim(left=min_date, right=max_date)
                ax_correction.set_ylim = [-100, self.project.MWAC*1.3*1000]
                ax_correction.legend()

                ax_cum_correction.set_title('POI Energy Correction')
                ax_cum_correction.plot(y_poi_cum_original, label='Original')
                ax_cum_correction.plot(y_poi_cum_corrected, label='Corrected')
                ax_cum_correction.legend()
            

        elif plot_name.replace(' ', '_').lower() == 'tracker':
            if self.project.Tracker_site:
                y = self.project.df[self.project.pos_Tracker]

                fig, ax = plt.subplots(num='Tracker Angles - {}'.format(self.project.project_name))
                fig.suptitle(self.project.project_name)

                ax.set_title('Tracker Angles')
                y.plot(ax=ax)
                ax.set_xlim(left=min_date, right=max_date)
                ax.set_ylim(-60,60)
        
        else:
            print('`{}` is not a valid plot name. See documentation for valid names & try again'.format(plot_name))
            fig, ax = 'Invalid plot_name', 'Invalid plot_name'


        # If we have multiple subplots we'll overwrite `ax` with the list of subplots stored in `axs`
        if not ax:
            try:
                ax = axs
            except:
                print('Error creating {} plot.'.format(plot_name))
                fig, ax = 'Error creating {} plot'.format(plot_name), 'Error creating {} plot'.format(plot_name)
        

        format_kwargs = {'screen': screen,
                         'fullscreen': fullscreen,
                         'tight_layout': tight_layout,
                         'default_tool': default_tool}

        if fig_list:
            for fig in fig_list:
                self.format_fig(fig, **format_kwargs)
        else:
            self.format_fig(fig, **format_kwargs)

        return Plot(fig, ax)


    def format_fig(self, fig, **kwargs):
        screen = kwargs.get('screen', None)
        fullscreen = kwargs.get('fullscreen', None)
        tight_layout = kwargs.get('tight_layout', None)
        default_tool = kwargs.get('default_tool', None)

        if isinstance(fig, Figure):
            try:
                # Move plots to the selected screen
                if screen:
                    if isinstance(screen, Number):
                        try:
                            screen = screen-1
                            screens = QGuiApplication.screens()
                            top_left = screens[screen].availableGeometry().topLeft()
                            fig.canvas.manager.window.move(top_left)
                        except IndexError:
                            warnings.warn('`screen` index out of range. Please supply a number less than or equal to the number of screens you have.')
                    else:
                        warnings.warn('TypeError - please supply a number for `screen` to select your desired plot screen')

                # Set plots to fullscreen if desired
                if fullscreen:
                    fig.canvas.manager.window.showMaximized()
                
                # Remove unnecessary whitespace around plots
                if tight_layout:
                    fig.set_tight_layout(True)

                # Set the default tool for zooming/panning
                if default_tool == 'zoom':
                    fig.canvas.manager.toolbar.zoom()
                elif default_tool == 'pan':
                    fig.canvas.manager.toolbar.pan()
            except Exception as e:
                print(e)

        
    def close_plots(self, plots=True):
        """Method to close plots"""
        if isinstance(plots, bool):
            plt.close('all')
        else:
            if isinstance(plots, str):
                plots = [plots]
            for plot_name in plots:
                self.close_plot(plot_name)

    
    def close_plot(self, plot_name):
        plot = self.plot_list[plot_name].fig
        plt.close(plot)



# -*- coding: utf-8 -*-
"""
Imports above docstring to exclude from Sphinx AutoAPI
Created on Wed Aug 16 09:38:55 2017

@author: Kevin Anderson
"""

import numpy as np
import os
import pandas as pd
import re
import smartsheet
from sqlalchemy import create_engine


username = os.path.split(os.path.expanduser("~"))[1]
if username == 'Kevin Anderson':
    # back in my day, C drives were good enough for anyone
    file_project = os.path.join(r'C:\Users', username, r'Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project')
elif username == 'blumenthal':
    file_project = r'C:\Users\blumenthal\Cypress Creek Renewables\AM-Performance - _Dashboard_Project'
    
elif username == 'Ryan':
    # the co-worker so nice, we elif'd him twice
    file_project = r'E:\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project'

elif username == 'MartinWaters':
    file_project = r'C:\Users\MartinWaters\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project'
    
elif username == 'EricFitch':
    # RIP
    file_project = r'C:\Users\EricFitch\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project'
    
elif username == 'ChristopherDowns':
    file_project=r'C:\Users\ChristopherDowns\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'
    
elif username == 'PerfEng':
    file_project = r'C:\Users\PerfEng\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'
    
elif username == 'corey.pullium':
    file_project = r'C:\Users\corey.pullium\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'
    
elif username == 'MelissaFrench':
    file_project = r'C:\Users\MelissaFrench\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project'

elif username == 'StoneHayden':
    file_project = r'C:\Users\StoneHayden\Cypress Creek Renewables\AM-Performance - _Dashboard_Project'
    
elif username == 'PradeepAmireddy':
    file_project = r'C:\Users\PradeepAmireddy\Cypress Creek Renewables\AM-Performance - _Dashboard_Project'
    
elif username == 'AnnaSchmackers':
    file_project = r'C:\Users\AnnaSchmackers\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'

elif username == 'LukeSain':
    file_project = r'C:\Users\LukeSain\OneDrive - Cypress Creek Renewables\Documents - AM-Performance\_Dashboard_Project'

elif username == 'AndrewCurthoys':
    file_project = r'C:\Users\AndrewCurthoys\Cypress Creek Renewables\AM-Performance - Documents\_Dashboard_Project'

else:
    # saurabh because he likes the D drive better. 
    # Wait...why do I have to be the 'else'. I demand to be the 'if'
    # dude you put yourself there, idk what you're complaining about
    # malarky! there is a plot against me. and I don't mean one of the dashboard variety.
    # typical asset-management performance engineer, playing the victim when he tilted the POA himself
    # RIP
    file_project = r'D:\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project'

file_production_data = os.path.split(file_project)[0]
    
def get_sql_engine():
    sql_username = os.environ['BARTERTOWN_USERNAME']
    sql_password = os.environ['BARTERTOWN_PASSWORD']
    encoding_dict = {
        ':': '%3A',
        '/': '%2F',
        '?': '%3F',
        '#': '%23',
        '[': '%5B',
        ']': '%5D',
        '@': '%40',
        '!': '%21',
        '$': '%24',
        '&': '%26',
        "'": '%27',
        '(': '%28',
        ')': '%29',
        '*': '%2A',
        '+': '%2B',
        ',': '%2C',
        ';': '%3B',
        '=': '%3D',
        '%': '%25',
        ' ': '%20',
    }
    pattern = re.compile('|'.join(re.escape(key) for key in encoding_dict.keys()))
    sql_password = pattern.sub(lambda x: encoding_dict[x.group()], sql_password)
    
    host = 'bartertown.cbnrsntwaejm.us-west-2.rds.amazonaws.com'
    port = '5432'
    db = 'thunderdome'
    
    url = 'postgresql+psycopg2://{}:{}@{}:{}/{}'
    url = url.format(sql_username, sql_password, host, port, db)
    
    engine = create_engine(url,pool_pre_ping=True)
    return engine

def get_ss_as_df(sheet_id, drop_col_1=True, start_col=None, index_col=None):
    # API_TOKEN is user-specific, so SS will block you from accessing 
    # sheets that you don't have permissions for
    API_TOKEN = os.environ['SMARTSHEET_TOKEN']

    # Initialize API client
    ss_client = smartsheet.Smartsheet(API_TOKEN)

    # by default, the SS functions will return error objects if you do something wrong 
    # (eg query a nonexistent sheet), instead of raising exceptions
    # if you ask me, that's crazy.  so we tell it to raise exceptions instead
    ss_client.errors_as_exceptions(True)

    cols = ss_client.Sheets.get_columns(sheet_id)
    col_ids, colnames = zip(*((col.id, col.title) for col in cols.result[start_col:]))

    # If an number is provided for the index col we'll find that index in the result list and assign it as the index
    if isinstance(index_col, (int, float)):
        index_col_id, index_colname = cols.result[index_col].id, cols.result[index_col].title
    # If a string is provided for the index col we'll find that column name & assign it as the index
    elif isinstance(index_col, str):
        for col in cols.result:
            if col.title == index_col:
                index_col_id, index_colname = col.id, col.title

    # Add index column to the col list if present
    try:
        col_ids = (index_col_id,) + col_ids # type: ignore
        colnames = (index_colname,) + colnames # type: ignore
    except NameError:
        pass
        
    # Get sheet
    sheet = ss_client.Sheets.get_sheet(sheet_id, column_ids=list(col_ids))
    
    # extract out the cell values into a df
    values = [[cell.value for cell in row.cells] for row in sheet.rows]
    df = pd.DataFrame(values)

    # Drop empty rows
    df = df.dropna(how='all')
    # Fill empty values with np.nan
    df = df.fillna(np.nan)
    
    # df has the data, but the column names are just numbers
    df.columns = list(colnames)

    # bring the index over
    df.index = range(len(df)) # type: ignore
    if drop_col_1:
        del df['Column1']
    
    # Set index if provided
    if index_col:
        df = df.set_index(index_colname) # type: ignore

    return df

def get_df_keys(retired=False):
    # df_keys SS ID
    sheet_id = '8659171076794244'
    all_df_keys = get_ss_as_df(sheet_id)
    all_df_keys = all_df_keys.set_index('Project')
    for col in ['PIS', 'FC']:
        all_df_keys[col] = pd.to_datetime(all_df_keys[col])
        
    if not retired:
        df_keys = all_df_keys.loc[all_df_keys['Retired'] != True, :]
    else:
        df_keys = all_df_keys

    return df_keys



# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
import statsmodels.api as sm

"""
Imports above docstring to exclude from Sphinx AutoAPI
Created on Wed Jul 27 08:25:53 2016

@author: SEBASTIAN
"""

def generate_Tcell(POA,Tamb,wind_speed, a,b, Delta_Tcnd):
    #  based on NREL formulas.
    Ref_Irrad = 1000.0
    Tmod = Tamb + POA *np.exp( a+ b*wind_speed)
    #Temp_drop = 3.0
    Tcell = Tmod + (POA / Ref_Irrad)*Delta_Tcnd
    return Tcell
#---    
def generate_Tcell_from_Tmod(POA,Tmod,Delta_Tcnd):
    Ref_Irrad = 1000.0
    #Temp_drop = 3.0
    Tcell = Tmod + (POA / Ref_Irrad)*Delta_Tcnd
    return Tcell

                       
def calculate_DC_Corrected_PVsyst(df, Pstc_KW , Gstc,Temp_Coeff_Pmax , Tcell_typ_avg, clipping_KW):
    # 
    #        Temperature corrected Theoretical DC energy over time (see NREL paper)
    #
    Generation = df['kWh_ORIGINAL']     #must ignore degradation for clipping filter to work, apply deg after
    POA = df['POA (W/m2)']
    
    df['DC_corrected_PVsyst'] = Pstc_KW * (POA.values / Gstc) *(1.0 - Temp_Coeff_Pmax*(Tcell_typ_avg - df['Tcell'].values)) 
    
    #
    #   Modify clipping points
    #   1) For PR, all DC_corrected is 0 and values wih POA < 100 = 0
    #   2)  For Weather adjusted, DC corrected is moved to the clipping point
    
    df['DC_corrected_PVsyst_PR'] = df['DC_corrected_PVsyst'].values 
    df['DC_corrected_PVsyst_WA'] = df['DC_corrected_PVsyst'].values
    df['Gen_NO_Clipping_PVsyst'] = Generation.values

    #
    #   Make 0 points that are 95% near clipping and POA < 100 W/m2
    #
    k_clipping = 0.98
    df.loc[(Generation > clipping_KW * k_clipping), ['DC_corrected_PVsyst_PR','Gen_NO_Clipping_PVsyst']] = 0.0
    #where dc power >0, and clipping has been assigned, get minimum of all dc_corrected values (the dc power clipping point)
    clipping_Point_DC_corrected_PVsyst = df.loc[(df['DC_corrected_PVsyst_PR'] == 0.0) & (df['DC_corrected_PVsyst'] > 0), 'DC_corrected_PVsyst_WA'].min()
    df.loc[(POA < 100.0), ['DC_corrected_PVsyst_PR','Gen_NO_Clipping_PVsyst']] = 0.0

    #
    #To calculate Weather Adjusted IE production, find smallest DC_Corrected when clipping
    #
    #clip = df[Generation > clipping_KW *k_clipping ]['DC_corrected_PVsyst'].min()
    #
    df.loc[ df['DC_corrected_PVsyst_WA'] > clipping_Point_DC_corrected_PVsyst , 'DC_corrected_PVsyst_WA'] = clipping_Point_DC_corrected_PVsyst
    df.loc[ df['DC_corrected_PVsyst_WA'] == clipping_Point_DC_corrected_PVsyst , ['DC_corrected_PVsyst_PR','Gen_NO_Clipping_PVsyst']] = 0
    ###df.loc[df['DC_corrected_PVsyst_WA'] > clipping_KW *k_clipping , 'DC_corrected_PVsyst_WA'] = clip

    #
    #df['Gen_NO_Clipping_PVsyst'] = Generation.values
    df.loc[Generation > clipping_KW * k_clipping , 'Gen_NO_Clipping_PVsyst'] = 0.0  
    #              
    return df             

def calculate_DC_Corrected(df, var_generation, Pstc_KW , Gstc,Temp_Coeff_Pmax , Tcell_typ_avg, clipping_Point_DC_corrected_PVsyst):
    # 
    #        Temperature corrected Theoretical DC energy over time (see NREL paper)
    #
    Generation = df[var_generation]
    POA = df['POA_avg']
    
    df['DC_corrected'] = Pstc_KW * (POA.values / Gstc) *(1.0 - Temp_Coeff_Pmax*(Tcell_typ_avg - df['Tcell'].values)) 
    
    #
    #   Modify clipping points
    #   1) For PR, all DC_corrected is 0 and values wih POA < 100 = 0
    #   2)  For Weather adjusted, DC corrected is moved to  the minimum point
    
    df['DC_corrected_PR'] = df['DC_corrected'].values 
    df['DC_corrected_WA'] = df['DC_corrected'].values
    df['Gen_NO_Clipping'] = Generation.values
    #
    #   Make 0 points that are 95% near clipping and POA < 100 W/m2
    #
    #k_clipping = 0.98
    df.loc[(POA < 100.0), ['DC_corrected_PR','Gen_NO_Clipping']] = 0.0
    df.loc[(df['DC_corrected'] > clipping_Point_DC_corrected_PVsyst), ['DC_corrected_PR','Gen_NO_Clipping']] = 0.0
    #
    #   To calculate Weather Adjusted IE production, find smallest DC_Corrected when clipping
    #
    clip =  clipping_Point_DC_corrected_PVsyst
    #                          
    df.loc[df['DC_corrected'] > clipping_Point_DC_corrected_PVsyst , 'DC_corrected_WA'] = clip
    #
    #df['Gen_NO_Clipping'] = Generation.values
    df.loc[df['DC_corrected'] > clipping_Point_DC_corrected_PVsyst  , 'Gen_NO_Clipping'] = 0.0  
    #              
    return df           

    
#-----
def generate_linear_coeff_table(df):
    df = df.set_index('date')
    df.index = pd.to_datetime(df.index)
    #
    df = df.astype('float')
    df['Clipping (KW)'] =  df['Clipping (W)'] /1000.0
    #  Because the way the data is in the IE report.  Clipping is removed from the Column of
    #  Actual production.
    df['Production_PVsyst (KW)'] = df['Year 0 Actual Production (kWh)'] + df['Clipping (KW)'] 
    #df.locdf['Year 0 Actual Production (kWh)'] + df['Clipping (KW)'] 
    #  Regression Model as defined in ASTM;  P = E (a1 + a2*E + a3*Ta +a4*v)  
    # Select variables
    power = 'Production_PVsyst (KW)'
    POA_i = 'POA (W/m2)'
    wind_speed ='Wind Velocity (m/s)'
    temp = 'Ambient Temperature'
    df_coef = pd.DataFrame(index = ['E', 'E2', 'ET', 'Ev'])   
    #    
    for i in range(1,13):
        aux = df[df.index.month ==i]
        #
        y = aux[power]
        X = aux[[POA_i, temp, wind_speed]]
        X = sm.add_constant(X, has_constant='add')
        X2 = X.multiply(aux[POA_i], axis = 'index')
        X2.columns = ['E', 'E2', 'ET', 'Ev']
        #
        
        
        
        
        
        
        
        
        
        res = sm.OLS(y, X2).fit()
        df_coef[str(i)] = res.params.values
    #
    df_coef.columns = ['Jan' , 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct','Nov', 'Dec']
    #df_coef.to_excel('Monthly_Linear_Reg_coef_CB_Bladen.xlsx')
    return df_coef    
    
def generate_linear_coeff_table_v2(df, var, clipping):
    #
    df[var] = df[var].astype('float')
    months = np.unique(df.index.month)
    #  Because the way the data is in the IE report.  Clipping is removed from the Column of
    #  Actual production.
    #df['Clipping (KW)'] =  df[var[1]] /1000.0
    #df['Production_PVsyst (KW)'] = df[var[0]] + df['Clipping (KW)'] 
    #df.locdf['Year 0 Actual Production (kWh)'] + df['Clipping (KW)'] 
    #  Regression Model as defined in ASTM;  P = E (a1 + a2*E + a3*Ta +a4*v)  
    # Select variables
    power = var[0]
    POA_i = var[1]
    wind_speed = var[2]
    temp = var[3]
    
    #filtering
    df = df.loc[(df[POA_i] > 400) & (df[POA_i] < 800), :]
    df = df.loc[(df[power] > 10) & (df[power] < clipping *.98), :]
    df = df.loc[(df[temp] > 0) & (df[temp] < 45), :]

    
    df_coef = pd.DataFrame(index = ['E', 'E2', 'ET', 'Ev']) 
    df_coef_RC = pd.DataFrame(index = ['POA', 'T', 'Wind'])
    #    
    for i in months:       
        aux = df[df.index.month ==i]
        #
        if not aux.empty:
            
            y = aux[power]
            aux['const'] = 1
            X = aux[['const', POA_i, temp, wind_speed]]
            #X = aux[[POA_i, temp, wind_speed]]
            #X = sm.add_constant(X)
            X2 = X.multiply(aux[POA_i], axis = 'index')
            X2.columns = ['E', 'E2', 'ET', 'Ev']
            #
            res = sm.OLS(y, X2).fit()
            df_coef[str(i)] = res.params.values
            df_coef_RC[str(i)] = [aux[POA_i].quantile(.6), aux[temp].mean(), aux[wind_speed].mean()]
        else:
            df_coef[str(i)] = [0]*4
            df_coef_RC[str(i)] = [0]*3
    #
    #df_coef.columns = ['Jan' , 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct','Nov', 'Dec']
    #df_coef.to_excel('Monthly_Linear_Reg_coef_CB_Bladen.xlsx')
    return df_coef, df_coef_RC   

def generate_linear_coeff_table_v3(df, var, clipping):
    #
    df[var] = df[var].astype('float')
    months = np.unique(df.index.month) #create np array for months
    
    #  Because the way the data is in the IE report.  Clipping is removed from the Column of
    #  Actual production.
    #df['Clipping (KW)'] =  df[var[1]] /1000.0
    #df['Production_PVsyst (KW)'] = df[var[0]] + df['Clipping (KW)'] 
    #df.locdf['Year 0 Actual Production (kWh)'] + df['Clipping (KW)'] 
    #  Regression Model as defined in ASTM;  P = E (a1 + a2*E + a3*Ta +a4*v)  
    # Select variables
    power = var[0]
    POA_i = var[1]
    wind_speed = var[2]
    temp = var[3]
    
    #filtering
    df = df.loc[(df[POA_i] > 400) & (df[POA_i] < 800), :]
    df = df.loc[(df[power] > 10) & (df[power] < clipping *.98), :]
    df = df.loc[(df[temp] > 0) & (df[temp] < 45), :]

    
    df_coef = pd.DataFrame(index = ['E', 'E2', 'ET', 'Ev']) 
    df_coef_RC = pd.DataFrame(index = ['POA', 'T', 'Wind'])
    #    
    for i in months:       
        aux = df[df.index.month ==i] #get the month for 8760
        #
        # need a reasonable number of points to do regression accurately after filtering above
        # use 10 as an arbitrary but reasonable limit
        #
        if not aux.empty and len(aux) > 10:
            
            y = aux[power] #create df of DV 
            aux['const'] = 1 #multiply POA in later on to isolate the value
            X = aux[['const', POA_i, temp, wind_speed]] #create dataframe of IVs
            #X = aux[[POA_i, temp, wind_speed]]
            #X = sm.add_constant(X)
            X2 = X.multiply(aux[POA_i], axis = 'index')
            X2.columns = ['E', 'E2', 'ET', 'Ev']
            #
            res = sm.OLS(y, X2).fit()
            df_coef[str(i)] = res.params.values #get coefficients in np array
            df_coef_RC[str(i)] = [aux[POA_i].quantile(.6), aux[temp].mean(), aux[wind_speed].mean()]
        else:
            df_coef[str(i)] = [0]*4
            df_coef_RC[str(i)] = [0]*3
    #
    #df_coef.columns = ['Jan' , 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct','Nov', 'Dec']
    #df_coef.to_excel('Monthly_Linear_Reg_coef_CB_Bladen.xlsx')
    return df_coef, df_coef_RC   


    
def create_ASTM_column (df, df_coef):
    df_coef.columns = range(1,13)
    df_coef = df_coef.T
    df_aux = pd.DataFrame(index = df.index, columns = df_coef.columns) 
    df_aux['month'] = df_aux.index.month
    # Generate monthly linear regression coeff.
    for i in range(1,13):
        df_aux.loc[df_aux['month'] == i , ['E']] = df_coef['E'][i]
        df_aux.loc[df_aux['month'] == i , ['E2']] = df_coef['E2'][i]
        df_aux.loc[df_aux['month'] == i , ['ET']] = df_coef['ET'][i]
        df_aux.loc[df_aux['month'] == i , ['Ev']] = df_coef['Ev'][i]
    
    P_exp = df['POA_avg'] * (df_aux['E'] + df_aux['E2']*df['POA_avg'] + df_aux['ET']*df['Tamb_avg'] + df_aux['Ev']*df['Wind_speed'])
    #  Calculate ASTM values for the Weatger corrected column.
    return P_exp    


from __future__ import annotations
from collections.abc import Callable
import functools
from numbers import Number
import pandas as pd
from pvlib.location import Location
import numpy as np
from scipy.interpolate import CubicSpline
from scipy.stats import gmean
from scipy.signal import find_peaks
from statistics import geometric_mean
from typing import TypedDict
from typing_extensions import Unpack

from ccrenew import DaylightParams

class SensorData(TypedDict):
    df: pd.DataFrame
    project_params: DaylightParams

def daylight(func: Callable):
    @functools.wraps(func)
    def calc_daylight_hours(**kwargs: Unpack[SensorData]):
        # Collect df & project to determine daylight hours
        df = kwargs.get('df')
        project_params = kwargs.get('project_params')

        # Calculate daylight hours
        lat = project_params.lat
        lon = project_params.lon
        tz = project_params.tz
        project_location = Location(lat, lon, tz)
        times = df.index.tz_localize(tz) # type: ignore
        df_suntimes = project_location.get_sun_rise_set_transit(times, method='spa')[['sunrise', 'sunset']]
        df_suntimes.index = pd.DatetimeIndex(df_suntimes.index)
        df = df.loc[df_suntimes.query("index.dt.hour >= sunrise.dt.hour and index.dt.hour <= sunset.dt.hour").index.tz_localize(None)] # type: ignore
        kwargs['df'] = df

        # Pass df with only daylight hours to the function
        result = func(**kwargs)
        return result
    return calc_daylight_hours

def comms(df: pd.DataFrame)-> pd.DataFrame:
    return df.isnull()

def zeroes(df: pd.DataFrame) -> pd.DataFrame:
    return df == 0

@daylight
def daylight_zeroes(*, df: pd.DataFrame, project_params: DaylightParams) -> pd.DataFrame:
    """
    Return zero values during daylight hours.

    Args:
        df (pd.DataFrame): Dataframe of project readings.
        project_params (DaylightParams): A namedtuple object with parameters for calculating daylight hours.

    Returns:
        pd.DataFrame A set of bool values based on zeroes.
    """
    return zeroes(df)

def frozen(df: pd.DataFrame, window: str|int|None = None) -> pd.DataFrame:
    """ 'cutoff_limit' is minimum amount of time values must be unchanged to be considered frozen. 
    This parameters can be provided in the ContractParameters class in the config script. 
    The default is to use the minimum timedelta found in the DAS data."""
    if not window:
        return (df != 0) & (df.diff() == 0)
    if isinstance(window, str):
        df_freq = pd.infer_freq(df.index) # type: ignore
        window_range = 2*pd.Timedelta(window) - pd.Timedelta(df_freq) # type: ignore
        window_range = window_range/df_freq # type: ignore
    elif isinstance(window, Number):
        window_range = window
    else:
        raise TypeError('cutoff_limit for frozen values must be a number representing window size or a string in format ##min')
    
    lookback = df.rolling(window=window_range).apply(lambda x: np.all(x==x[0]), raw=True)
    forward_indexer = pd.api.indexers.FixedForwardWindowIndexer(window_size=window_range)
    lookforward = df.rolling(window=forward_indexer).apply(lambda x: np.all(x==x[0]), raw=True)
    frozen = lookback.fillna(0) + lookforward.fillna(0)

    return frozen > 0

def frozen_center(df: pd.DataFrame, window: str|int|None = None) -> pd.DataFrame:
    """ 'cutoff_limit' is minimum amount of time values must be unchanged to be considered frozen. 
    This parameters can be provided in the ContractParameters class in the config script. 
    The default is to use the minimum timedelta found in the DAS data.  """
    if not window:
        return (df != 0) & (df.diff() == 0)
    if isinstance(window, str):
        df_freq = pd.infer_freq(df.index) # type: ignore
        window_range = 2*pd.Timedelta(window) - pd.Timedelta(df_freq) # type: ignore
    elif isinstance(window, Number):
        window_range = window
    else:
        raise TypeError('cutoff_limit for frozen values must be a number representing window size or a string in format ##min')

    frozen = df.rolling(window_range, center=True).apply(lambda x: np.all(x == x[0]), raw=True) # type: ignore

    return frozen == 1

@daylight
def daylight_frozen(*, df: pd.DataFrame, project_params: DaylightParams, window: str|int|None = None) -> pd.DataFrame:
    return frozen(df, window)

def negatives(df: pd.DataFrame, cols: list=[]) -> pd.DataFrame:
    df = df.copy()
    if cols:
        df = df[cols] < 0
    else:
        df = df <0
    return df

def decreasing(df: pd.DataFrame) -> pd.DataFrame:
    return df.diff() < 0

def band_pass(df: pd.DataFrame, col_limits: list[tuple]) -> pd.DataFrame:
    df=df.copy()
    for col, limits in col_limits:
        df.loc[:,col] = (df[col].lt(limits[0])) | (df[col].gt(limits[1]))

    return df

def poa_mistracking(df: pd.DataFrame, degree: int=8) -> pd.DataFrame:
    df = df.copy()
    for day, df_daily in df.groupby(df.index.date): # type: ignore
        x = df_daily.index.hour

        # Find peaks for Solcast data
        y_solcast_raw = df_daily['Solcast']
        y_solcast = y_solcast_raw.fillna(0)
        fit_solcast = np.poly1d(np.polyfit(x, y_solcast, degree))

        x_fit = np.linspace(x[0], x[-1], 100)
        y_fit_solcast = fit_solcast(x_fit)

        # Count the number of hours that are above half the maximum value for the series
        # Trackers will have more hours above this value than non-trackers
        count_max_50_solcast = (y_solcast>y_solcast_raw.max()/2).sum()
        max_66_solcast = y_solcast_raw.max()*2/3

        # Only consider peaks that are at least 2/3rds as high as the maximum
        peaks_solcast, _ = find_peaks(y_fit_solcast, height=max_66_solcast)
        peak_count_solcast = len(peaks_solcast)


        # Loop through POA columns to find peaks
        for col in df.columns[df.columns != 'Solcast']:
            y_raw = df_daily[col]
            y = y_raw.fillna(0)
            fit = np.poly1d(np.polyfit(x, y, degree))

            x_fit = np.linspace(x[0], x[-1], 100)
            y_fit = fit(x_fit)

            # Calculate % of max values for use in calculations
            count_max_50 = (y_raw>y_raw.max()/2).sum()
            max_66 = y_raw.max()*2/3

            # Only consider peaks that are at least 2/3rds as high as the maximum
            peaks, _ = find_peaks(y_fit, height=max_66)
            peak_count = len(peaks)

            # If only 1 peak & Solcast shows more than 1, we'll consider it mistracking
            # If the number of points above 50% max for Solcast vs Native is 2 or more, we'll consider it mistracking
            if peak_count < 2 and peak_count_solcast >= 2:
                df.loc[df.index.date==day, col] = True # type: ignore
            elif count_max_50_solcast - count_max_50 > 1:
                df.loc[df.index.date==day, col] = True # type: ignore
            else:
                df.loc[df.index.date==day, col] = False # type: ignore

    df = df.drop(columns='Solcast').astype(bool)

    return df

@daylight
def daylight_poa_mistracking(*, df: pd.DataFrame, project_params: DaylightParams, degree: int=8):
    return poa_mistracking(df, degree)

def spline_filter(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy()
    df_bool = ~df.isna()
    for col in df.columns:
        non_nulls = df[col][df[col].notna()]
        x = non_nulls.index.values.astype('float')
        y = non_nulls.values
        cs = CubicSpline(x, y)
        deriv = cs(x, 1)
        deriv_mean = np.mean(deriv)
        stdev = np.std(deriv)
        y = (deriv < -2.5*stdev) | (deriv > 2.5*stdev)
        bool_col = pd.DataFrame(index=pd.to_datetime(x), data={col: y})
        df_bool.update(bool_col)
    df_bool = df_bool.astype(bool)

    return df_bool


if __name__ == '__main__':
    dates = pd.date_range('2023-1-1 00:00', '2023-6-1 00:00', freq='5T')
    values = list(range(len(dates)))

    data = np.array([values, values, values, values, values]).T
    df = pd.DataFrame(index=dates, data=data)
    df.loc[df.index < '2023-01-01 00:10'] = 1
    df.loc[(df.index > '2023-01-01 00:20') & (df.index < '2023-01-01 01:00')] = 1
    df.loc[(df.index > '2023-01-01 01:10') & (df.index < '2023-01-01 01:40')] = -1
    s = df.iloc[:,[0]]
    frozen(s, '15T')


#!/usr/bin/env python

"""
Imports above docstring to exclude from Sphinx AutoAPI
API for modelling solar+storage dispatch over shorter timer periods.
"""

__author__ = "Allen Lawrence"
__credits__ = "Brian Knowles", "Michael Baker"
__version__ = "0.1.0"
__maintainer__ = "Allen Lawrence"
__email__ = "allen.lawrence@ccrenew.com"
__status__ = "Development"


import numpy as np

### BINARY LOGIC ###
def binDischarge(cue, batt_power, soc, inp, lim, discharge_eff):
    """
    Calculates battery discharge for a single time step.
    Adapted from Brian Knowles's (CCR) battery model.

    :param cue: binary to cue to signal discharge/charge (0 or 1)
    :param batt_power: max energy per time interval dischargeable or chargeable from the battery (from the perspective
        of the grid).
    :param soc: current state-of-charge of battery
    :param inp: current solar production (energy per time step)
    :param lim: the goal power
    :param discharge_eff: efficiency of battery discharge (fraction).
    :return: scalar discharge value.
    """
    if cue:
        return min(batt_power/discharge_eff, soc, max(lim - inp, 0))
    else:
        return 0


def binCharge(cue, inp, lim, charge_eff, batt_power, batt_energy, soc):
    """
    Calculates battery charge for a single time step.
    Adapted from Brian Knowles's (CCR) battery model.

    :param cue: binary to cue to signal discharge/charge (0 or 1)
    :param inp: current solar production (energy per time step)
    :param lim: the goal power
    :param charge_eff: efficiency of battery charge (fraction)
    :param batt_power: max energy per time interval dischargeable or chargeable from the battery (from the perspective
        of the grid).
    :param batt_energy: maximum state-of-charge of battery (kWh)
    :param soc: current state-of-charge
    :return: scalar charge value.
    """
    if cue:
        return min(max(charge_eff * (inp - lim), 0), charge_eff * batt_power, batt_energy - soc)
    else:
        return min(charge_eff * inp, charge_eff * batt_power, batt_energy - soc)
####################




class Battery(object):
    """
    Battery model. Contains methods for battery dispatch.

    :param self.limit: scalar battery goal power. This can be overwritten by a variable signal in dispatch methods.
    :param self.power: battery charge/discharge power.
    :param self.duration: amount of time during which battery can discharge at max power.
    :param self.charge_eff: efficiency of battery charge.
    :param self.discharge_eff: efficiency of battery discharge.
    :param self.energy: maximum stored energy (from battery perspective)
    """

    def __init__(self):

        self.limit = 0.
        self.power = 0.
        #self.duration = 0.
        self.charge_eff = 0.85
        self.discharge_eff = 1.0
        self.nameplate_duration = 0.
        # degradation hours is the number of hours since the ESUs
        # were PIS'd.  since batteries don't know about history, 
        # this needs to be set to an appropriate value for previous
        # charge/discharge cycling
        self.degradation_hours = 0.0
        # calculate hourly degradation rate from the 4% annual figure
        self.degradation_rate = 0.96**(1/8760.0)

    @property
    def duration(self):
        deg_coeff = min(self.degradation_rate ** self.degradation_hours, 1)
        return self.nameplate_duration * deg_coeff

    @property
    def energy(self):
        return (self.power*self.duration)/self.discharge_eff



    def binary(self, valueS, inpS, limitS=False, soc0=0.):
        """
        Dispatch method for discharging "on-peak" and charging "off-peak". Interprets "on" vs. "off-peak" from the
            value signal provided.

        :param valueS: 1D nump.ndarray time series of energy values. Typically in units of $/kWh. If this signal
            contains more than two different values, consider using another dispatch method.
        :param inpS: 1D numpy.ndarray time series of energy production. Typically in units of kWh. If production
            clipping is significant, consider using another dispatch method.
        :param limitS: 1D numpy.ndarray time series of goal power. The battery will discharge to try to meet these
            values, and charge with input energy in excess of these values.
        :param soc0: Scalar value of internal battery energy. This is the SOC at the start of the dispatch period.
        :return: dict of output signals.
        """

        n = inpS.shape[0]
        assert valueS.shape[0] == n

        if not limitS:
            limitS = self.limit * np.ones(n)
        else:
            assert limitS.shape[0]== n

        #### BUILD  CUES FROM VALUE SIGNAL ####
        cueS = np.zeros(n)
        cueS[valueS == valueS.max()] = 1

        socS = np.zeros(n+1)
        socS[0] = soc0
        dischargeS = np.zeros(n)
        chargeS = np.zeros(n)

        for i in range(n):
            self.degradation_hours += 1
            soc = socS[i]
            cue = cueS[i]
            inp = inpS[i]
            lim = limitS[i]
            dischargeS[i] = binDischarge(cue, self.power, soc, inp, lim, self.discharge_eff)
            chargeS[i] = binCharge(cue, inp, lim, self.charge_eff, self.power, self.energy, soc)
            socS[i + 1] = soc + chargeS[i] - dischargeS[i]


        g_chargeS = chargeS / self.charge_eff
        g_dischargeS = dischargeS * self.discharge_eff
        batt_outS = inpS - g_chargeS + g_dischargeS

        return {'solar+storage': batt_outS,
                'charge': g_chargeS,
                'discharge': g_dischargeS,
                'SOC': socS
                }




#    def perfect(self, valueS, inpS, limitS=False, soc0=0.):
#        #TODO: discuss with O&M how to handle edge effects... add soc1 parameter?
#        """
#        Dispatch method for maximizing value of solar+storage production signal, given value signal and production
#            signal of any form. Employs GLPK's simplex algorithm.
#
#        Because this method only covers one optimization period, it will neglect optimal preparation for the next
#            period. If you don't understand what I'm talking about, contact the maintainer.
#
#        :param valueS: 1D nump.ndarray time series of energy values. Typically in units of $/kWh.
#        :param inpS: 1D numpy.ndarray time series of energy production. Typically in units of kWh.
#        :param limitS: 1D numpy.ndarray time series of goal power. The battery will discharge to try to meet these
#            values, and charge with input energy in excess of these values.
#        :param soc0: Scalar value of internal battery energy. This is the SOC at the start of the dispatch period.
#        :return: dict of output signals.
#        """
#
#        n = inpS.shape[0]
#        assert valueS.shape[0] == n
#
#        if not limitS:
#            limitS = self.limit * np.ones(n)
#        else:
#            assert limitS.shape[0] == n
#
#
#        ######### CONSTRAINT MATRICES #########
#        # Only charge from solar production
#        A = np.zeros((7 * n, 3 * n))
#        for i in range(n):
#            #### SOC #####
#            A[i][0:i + 1] = -1; A[i][n:n + i + 1] = 1; A[i][2 * n:2 * n + i + 1] = 1
#            A[n + i][0:i + 1] = 1; A[n + i][n:n + i + 1] = -1; A[n + i][2 * n:2 * n + i + 1] = -1
#            #### DISCHARGE ####
#            A[2 * n + i][i] = 1
#            A[3 * n + i][i] = 1
#            #### PROCHARGE + CHARGE #######
#            A[4 * n + i][n + i] = 1; A[4 * n + i][2 * n + i] = 1
#            A[5 * n + i][n + i] = 1; A[5 * n + i][2 * n + i] = 1
#            #### PROCHARGE ######
#            A[6 * n + i][n + i] = 1
#        A = np.asmatrix(A)
#
#
#        dischargeS, chargeS, socS = optiBatt(A, n, valueS, inpS, limitS, self.power, self.energy, self.discharge_eff,
#                                              self.charge_eff, soc0)
#
#        g_chargeS = chargeS / self.charge_eff
#        g_dischargeS = dischargeS * self.discharge_eff
#        batt_outS = inpS - g_chargeS + g_dischargeS
#
#        return {'solar+storage': batt_outS,
#                'charge': g_chargeS,
#                'discharge': g_dischargeS,
#                'SOC': socS
#                }


    def forecast(self):
        """
        IN DEVELOPMENT
        Similar to perfect, but where scheduling and execution are different, due to forecast uncertainty.
        """
        pass







"""Package for working with Cypress Creek Renewables Dashboard process
"""
import warnings

from pandas.errors import PerformanceWarning
warnings.simplefilter(action='ignore', category=(FutureWarning, PerformanceWarning)) # type: ignore
warnings.filterwarnings(action='ignore', message='divide by zero')

from collections import namedtuple


# Create namedtuple instances to store information to pass between modules
Colmap = namedtuple('Colmap', ['col_list', 'col_avg'])


print('Importing DashboardSession')
from ccrenew import (
    all_df_keys,
    cloud_data
)
from ccrenew.dashboard.utils.dashboard_utils import func_timer
from ccrenew.data_determination import daylight
from ccrenew.dashboard.plotting.plots import Plotter
from ccrenew.dashboard.session import DashboardSession
from ccrenew.dashboard.project import Project
from ccrenew.dashboard.utils.project_neighbors import (
    find_nearby_projects,
    find_nearby_similar_projects,
    find_similar_projects
)


# Helper functions
class Helpers:
    def __init__(self):
        self.df_keys = all_df_keys
        self.find_nearby_projects = find_nearby_projects
        self.find_nearby_similar_projects = find_nearby_similar_projects
        self.find_similar_projects = find_similar_projects
        self.get_sat_weather = cloud_data.get_sat_weather
        self.daylight = daylight

helpers = Helpers()


from concurrent.futures import (
    as_completed,
    ProcessPoolExecutor)
from functools import partial
import os
import time


def process_pool(kwargslist):

    # We'll use half of the available processors for the machine to not completely bog it down.
    max_workers = int(os.cpu_count()/2)

    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        # We'll set up `process_batch` as a partial so it will execute in parallel for each set of kwargs in the kwargslist
        process_project = partial(_process_batch)
        futures = [executor.submit(process_project, **kwargs) for kwargs in kwargslist]
        for future in as_completed(futures):
            result = future.result()
            project_name = result['project_name']
            error_info = result['error_info']
            project_num = result['project_num']
            project_total = result['project_total']
            project_time = result['project_time']
            batch_time = result['batch_time']
            if error_info:
                print('*************************************************************************************************************************')
                print("***** ERROR *****")
                print("***** ERROR *****")
                print(f'    Error encountered while processing "{project_name}". Project #{project_num} of {project_total}. Error details: {error_info}')
                print("***** ERROR *****")
                print("***** ERROR *****")
                print('*************************************************************************************************************************')
            else:
                print('*************************************************************************************************************************')
                print(f"Completed {project_name}. Project #{project_num} of {project_total} processed. Time to process: {project_time:.2f}s. Total elapsed time: {batch_time:.2f}s")
                print('*************************************************************************************************************************')
        
        return futures


def _process_batch(**kwargs):
    project_start = time.time()

    # Collect kwargs
    session = kwargs.get('session')
    data_source = kwargs.get('data_source')
    project_name = kwargs.get('project_name')
    reprocess = kwargs.get('reprocess')
    project_num = kwargs.get('project_num')
    project_total = kwargs.get('project_total')
    batch_start = kwargs.get('batch_start')

    # Initialize new session
    session = session(data_source=data_source)

    # Process project
    project = session.process_project(project_name=project_name, reprocess=reprocess)
    project_time = time.time()-project_start
    batch_time = time.time()-batch_start

    # Return dict for parsing the results
    return {'project_list': session.project_list.items(),
            'project_name': project_name,
            'project_num': project_num,
            'error_info': project.error_info,
            'project_total': project_total,
            'project_time': project_time,
            'batch_time': batch_time
    }


def export_pool(kwargslist):

    # We'll use half of the available processors for the machine to not completely bog it down.
    max_workers = int(os.cpu_count()/2)

    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        # We'll set up `process_batch` as a partial so it will execute in parallel for each set of kwargs in the kwargslist
        export_project = partial(_export_batch)
        futures = [executor.submit(export_project, **kwargs) for kwargs in kwargslist]
        for future in as_completed(futures):
            # Collect results for updating progress
            result = future.result()
            project_name = result['project_name']
            error_info = result['error_info']
            project_num = result['project_num']
            project_total = result['project_total']
            project_time = result['project_time']
            batch_time = result['batch_time']

            if error_info:
                print('*************************************************************************************************************************')
                print("***** ERROR *****")
                print("***** ERROR *****")
                print(f"    Error encountered while exporting {project_name}. Project #{project_num} of {project_total}. Error details: {error_info}")
                print("***** ERROR *****")
                print("***** ERROR *****")
                print('*************************************************************************************************************************')
            else:
                print('*************************************************************************************************************************')
                print(f"Completed {project_name}. Project #{project_num} of {project_total} exported. Time to export: {project_time:.2f}s. Total elapsed time: {batch_time:.2f}s")
                print('*************************************************************************************************************************')
        
        return futures


def _export_batch(**kwargs):
    project_start = time.time()

    # Collect kwargs
    session = kwargs.get('session')
    data_source = kwargs.get('data_source')
    project_name = kwargs.get('project_name')
    func_args = kwargs.get('func_args')
    project_num = kwargs.get('project_num')
    project_total = kwargs.get('project_total')
    batch_start = kwargs.get('batch_start')

    # Initialize new session
    session = session(data_source=data_source)

    # Process project
    error_info = ''
    try:
        session.export_project(project_name=project_name, **func_args)
    except Exception as e:
        error_info = e

    project_time = time.time()-project_start
    batch_time = time.time()-batch_start

    # Return dict for parsing the results
    return {'project_list': session.project_list.items(),
            'project_name': project_name,
            'error_info': error_info,
            'project_num': project_num,
            'project_total': project_total,
            'project_time': project_time,
            'batch_time': batch_time
    }


"""
Module to calculate distances between projects & acceptable neighbors to use
for data substitution
"""

import logging
import numpy as np
from pandas import DataFrame, Series

from ccrenew import (
    all_df_keys as df_keys,
    Numeric,
    SeriesLike)

# logger = logging.getLogger(__name__)


def find_nearby_projects(project_name: str, dist:int = 10, print_data:bool = True,
                         include_retired:bool = False, df:DataFrame = DataFrame()):
    """Creates a list of sites within a certain distance of the reference site.

    Args:
        project_name (str): Reference site.
        dist (int, optional): Number of miles to search around the reference site.
            Defaults to 10.
        include_retired (bool, optional): Whether to return sites that have been retired.
            Defaults to False, i.e. does not return retired sites.
        df (Dataframe, optional): Dataframe containing sitenames & lat/longs.
            Defaults to `df_keys` if no `df` supplied.

    Returns:
        nearby_sites (DataFrame): Contains information about all sites within the specified\
        distance from the reference site.
    """
    # Set `df` to `df_keys` if not supplied
    if df.empty:
        df = df_keys

        # Exclude projects with ".1" in the CCR_ID
        exclude_projects = df[['CCR_ID']].fillna('0').query("~CCR_ID.str.contains('\\.1')")
        df = df.loc[exclude_projects.index]

        # Exclude retired projects if desired
        if not include_retired:
            df = df.query("Retired != True")

    # Rename 'Tilt/GCR' column to a valid python identifier
    df = df.copy().rename(columns={'Tilt/GCR':'Tilt_GCR'})
    
    # Get a dictionary of site information, then the required parameters for the query
    site_info = df.loc[[project_name]].to_dict('index')[project_name]
    lat = site_info['GPS_Lat']
    lon = site_info['GPS_Long']
    df.loc[:, 'dist'] = haversine(lon, lat, df['GPS_Long'], df['GPS_Lat'])
    nearby_sites = df.query("dist <= @dist") \
                     .reindex(columns=['Racking', 'OEM_Racking', 'Tilt_GCR', 'Max_angle', 'Temp_Coeff_Pmax', 'dist']) \
                     .sort_values('dist')

    if print_data:
        print('\n')
        print('Sites within {} miles of {}'.format(dist, project_name))
        print(nearby_sites)
        print('\n')
    return nearby_sites

def find_similar_projects(project_name: str, include_retired:bool = False, df:DataFrame = DataFrame()):
    """Creates a list of sites that share similar racking properties to the reference project.

    Args:
        project_name (str): Reference project.
        include_retired (bool, optional): Whether to return sites that have been retired.
            Defaults to False, i.e. does not return retired sites.
        df (Dataframe, optional): Dataframe containing sitenames & lat/longs.
            Defaults to `df_keys` if no `df` supplied.
    Returns:
        DataFrame: Contains information about all sites with similar racking\
        properties to the reference site.
    """
    # Set `df` to `df_keys` if not supplied
    if df.empty:
        df = df_keys

        # Exclude projects with ".1" in the CCR_ID
        exclude_projects = df[['CCR_ID']].fillna('0').query("~CCR_ID.str.contains('\\.1')")
        df = df.loc[exclude_projects.index]

        # Exclude retired projects if desired
        if not include_retired:
            df = df.query("Retired != True")

    # Rename 'Tilt/GCR' column to a valid python identifier
    df = df.rename(columns={'Tilt/GCR':'Tilt_GCR'})
    
    # Get a dictionary of site information, then the required parameters for the query
    site_info = df.loc[[project_name]].to_dict('index')[project_name]
    racking = site_info['Racking']
    oem_racking = site_info['OEM_Racking']
    tilt = site_info['Tilt_GCR']
    temp_coeff = site_info['Temp_Coeff_Pmax']
    temp_coeff_cutoff = -0.39
    max_angle = site_info['Max_angle']
    
    if racking == 'Fixed':
        query = """Racking == @racking \
                 & Tilt_GCR == @tilt
                """

    elif racking == 'Tracker':
        # Correct for cases when GCR is entered as a percentage rather than a fraction
        if tilt > 1:
            tilt /= 100.0
        query = """Racking == @racking \
                 & Tilt_GCR >= @tilt - 0.05 \
                 & Tilt_GCR <= @tilt + 0.05 \
                 & Max_angle == @max_angle \
                 & (Temp_Coeff_Pmax - @temp_coeff_cutoff <= 0) == (@temp_coeff - @temp_coeff_cutoff <= 0)
                """
    
    else:
        print("Racking type not found for {}. Double check df_keys to ensure racking is either 'Fixed' or 'Tracker'".format(project_name))
        query = "index == @project_name"

    # Get all sites that have matching info to the reference site
    similar_sites = df.query(query)

    return similar_sites

def find_nearby_similar_projects(project_name:str, dist:int = 10, print_data:bool = True, include_retired:bool = False, df:DataFrame = DataFrame()):
    """Creates a list of sites within a certain distance of the reference site
    that share similar racking properties. 

    Args:
        project_name (str): Reference site.
        dist (int, optional): Number of miles to search around the reference site.
            Defaults to 10.
        include_retired (bool, optional): Whether to return sites that have been retired.
            Defaults to False, i.e. does not return retired sites.
        df (Dataframe, optional): Dataframe containing sitenames & lat/longs.
            Defaults to `df_keys` if no `df` supplied.
    Returns:
        DataFrame: Contains information about all sites within a certain distance\
        of the reference site that share similar racking properties.
    """
    nearby_sites = find_nearby_projects(project_name, print_data=False, dist=dist, include_retired=include_retired, df=df)
    nearby_similar_sites = find_similar_projects(project_name, df=nearby_sites)
    
    if print_data:
        print('\n')
        print('Similar sites within {} miles of {}'.format(dist, project_name))
        print(nearby_similar_sites)
        print('\n')
    return nearby_similar_sites

def haversine(lon1: Numeric|SeriesLike, lat1: Numeric|SeriesLike, lon2: Numeric|SeriesLike, lat2: Numeric|SeriesLike) -> Numeric|SeriesLike:
    """Calculate the great circle distance between two points
    on the earth (specified in decimal degrees).

    Args:
        lon1 (Number, np.ndarray, or Series): Longitude of point 1.
        lon1 (Number, np.ndarray, or Series): Latitude of point 1.
        lon1 (Number, np.ndarray, or Series): Longitude of point 2.
        lon1 (Number, np.ndarray, or Series): Latitude of point 2.

    Returns:
        np.number, np.ndarray, or Series: Distance between points in miles.
    """
    lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = np.sin(dlat/2.0)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon/2.0)**2

    c = 2 * np.arcsin(np.sqrt(a))
    dist_miles = 3959 * c
    return dist_miles